from typing import Any, Dict, List, Optional, Tuple

from powl.discovery.total_order_based.inductive.variants.brute_force.factory import (
    CutFactoryPOWLBruteForce,
)
from powl.discovery.total_order_based.inductive.variants.im_tree import IMBasePOWL, T
from powl.discovery.total_order_based.inductive.variants.powl_discovery_varaints import (
    POWLDiscoveryVariant,
)
from powl.objects.obj import POWL


class POWLInductiveMinerBruteForce(IMBasePOWL):
    def instance(self) -> POWLDiscoveryVariant:
        return POWLDiscoveryVariant.BRUTE_FORCE

    def find_cut(
        self, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[Tuple[POWL, List[T]]]:
        res = CutFactoryPOWLBruteForce.find_cut(obj, parameters=parameters)
        return res



from enum import auto, Enum


class POWLDiscoveryVariant(Enum):
    TREE = auto()  # base IM with no partial orders
    BRUTE_FORCE = auto()
    MAXIMAL = auto()
    DYNAMIC_CLUSTERING = auto()
    DECISION_GRAPH_MAX = auto()
    DECISION_GRAPH_CLUSTERING = auto()
    DECISION_GRAPH_CYCLIC = auto()
    DECISION_GRAPH_CYCLIC_STRICT = auto()



from typing import Any, Dict

import pm4py
from pm4py.objects.ocel.obj import OCEL

from powl.discovery.object_centric.variants.oc_powl.utils.divergence_free_graph import (
    get_divergence_free_graph,
)
from powl.discovery.object_centric.variants.oc_powl.utils.filtering import (
    keep_most_frequent_activities,
)
from powl.discovery.object_centric.variants.oc_powl.utils.interaction_properties import (
    get_interaction_patterns,
)
from powl.discovery.object_centric.variants.oc_tree.utils.oc_process_trees import (
    load_from_pt,
)
from powl.discovery.object_centric.variants.oc_tree.utils.tree_to_ocpn_conversion import (
    convert_ocpt_to_ocpn,
)


def apply(
    oc_log: OCEL,
    activity_coverage_threshold: float = 1.0,
) -> Dict[str, Any]:

    relations = oc_log.relations
    relations = keep_most_frequent_activities(
        relations, coverage=activity_coverage_threshold
    )

    div, con, rel, defi = get_interaction_patterns(relations)
    df2_graph, _ = get_divergence_free_graph(relations, div, rel)

    tree = pm4py.discover_process_tree_inductive(df2_graph)
    ocpt = load_from_pt(tree, rel, div, con, defi)
    ocpn = convert_ocpt_to_ocpn(ocpt)

    return ocpn



from abc import ABC
from collections import Counter
from itertools import combinations
from typing import Any, Collection, Dict, Generic, List, Optional, Tuple

from pm4py.algo.discovery.inductive.cuts.abc import Cut, T
from pm4py.algo.discovery.inductive.dtypes.im_dfg import InductiveDFG
from pm4py.algo.discovery.inductive.dtypes.im_ds import (
    IMDataStructureDFG,
    IMDataStructureUVCL,
)
from pm4py.objects.dfg import util as dfu
from pm4py.objects.dfg.obj import DFG
from pm4py.objects.dfg.util import get_transitive_relations

from pm4py.algo.discovery.inductive.variants.imf import IMFParameters
from pm4py.util import exec_utils
from powl.discovery.total_order_based.inductive.utils.filtering import (
    FILTERING_THRESHOLD,
    FILTERING_TYPE,
    FilteringType,
)

from powl.general_utils.efg_frequency_filtering import filter_efg_based_on_filtered_dfg
from powl.objects.BinaryRelation import BinaryRelation
from powl.objects.obj import POWL, StrictPartialOrder


def generate_initial_order(nodes, efg):
    po = BinaryRelation(nodes)
    for a, b in combinations(nodes, 2):
        if (a, b) in efg:
            if not (b, a) in efg:
                po.add_edge(a, b)
        else:
            if (b, a) in efg:
                po.add_edge(b, a)
    return po


def remove(blocks, g):
    res = []
    for g2 in blocks:
        if not g2.__str__().__eq__(g.__str__()):
            res.append(g2)
    return res


def contains(blocks, g):
    for g2 in blocks:
        if g2.__str__().__eq__(g.__str__()):
            return True
    return False


def is_valid_order(po, efg, start_activities, end_activities):
    if po is None:
        return False

    if len(po.nodes) < 2:
        return False

    if not po.is_strict_partial_order():
        return False

    start_blocks = po.nodes
    end_blocks = po.nodes

    for group_1, group_2 in combinations(po.nodes, 2):

        edge_g1_g2 = po.is_edge(group_1, group_2)
        edge_g2_g1 = po.is_edge(group_2, group_1)

        if edge_g1_g2:
            start_blocks = remove(start_blocks, group_2)
            end_blocks = remove(end_blocks, group_1)
        if edge_g2_g1:
            start_blocks = remove(start_blocks, group_1)
            end_blocks = remove(end_blocks, group_2)

        all_ef_g1_g2 = True
        all_ef_g2_g1 = True

        for a in group_1:
            for b in group_2:
                if not (a, b) in efg:
                    all_ef_g1_g2 = False
                if not (b, a) in efg:
                    all_ef_g2_g1 = False
        if all_ef_g1_g2 and all_ef_g2_g1 and (edge_g1_g2 or edge_g2_g1):
            return False
        if not edge_g1_g2 and not edge_g2_g1 and not (all_ef_g1_g2 and all_ef_g2_g1):
            return False

    n = len(po.nodes)
    for i in range(n):
        group = po.nodes[i]
        c1 = contains(start_blocks, group)
        c2 = len(set(group).intersection(start_activities)) > 0
        c3 = contains(end_blocks, group)
        c4 = len(set(group).intersection(end_activities)) > 0
        if (c1 and not c2) or (c3 and not c4):
            return False

    return True


def cluster_order(binary_relation):
    pre = {node: set() for node in binary_relation.nodes}
    post = {node: set() for node in binary_relation.nodes}
    for node1 in binary_relation.nodes:
        for node2 in binary_relation.nodes:
            if binary_relation.is_edge(node1, node2):
                pre[node2].add(node1)
                post[node1].add(node2)

    clusters = []
    for node in binary_relation.nodes:
        matched = False
        for i in range(len(clusters)):
            cluster = clusters[i]
            if pre[node] == pre[cluster[0]] and post[node] == post[cluster[0]]:
                clusters[i].append(node)
                matched = True
                break
        if not matched:
            clusters.append([node])

    new_relation = BinaryRelation([tuple(c) for c in clusters])
    for cluster1 in new_relation.nodes:
        for cluster2 in new_relation.nodes:
            node1 = cluster1[0]
            node2 = cluster2[0]
            if binary_relation.is_edge(node1, node2):
                new_relation.add_edge(cluster1, cluster2)

    return new_relation


def get_efg(post_sets):
    efg = Counter()
    for a, post_set in post_sets.items():
        for b in post_set:
            efg[(a, b)] += 1
    return efg


class MaximalPartialOrderCut(Cut[T], ABC, Generic[T]):
    @classmethod
    def operator(
        cls, parameters: Optional[Dict[str, Any]] = None
    ) -> StrictPartialOrder:
        return StrictPartialOrder([])

    @classmethod
    def holds(
        cls, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[BinaryRelation]:

        dfg = obj.dfg
        alphabet = sorted(dfu.get_vertices(dfg), key=lambda g: g.__str__())
        noise_threshold = None
        if FILTERING_TYPE in parameters.keys():
            filtering_type = parameters[FILTERING_TYPE]
            if filtering_type is FilteringType.DFG_FREQUENCY:
                noise_threshold = exec_utils.get_param_value(
                    IMFParameters.NOISE_THRESHOLD, parameters, 0.0
                )
                print('FILTERING: ', noise_threshold)

        if type(obj) is IMDataStructureUVCL:
            efg = filter_efg_based_on_filtered_dfg(obj, alphabet, dfg, noise_threshold)
        elif type(obj) is IMDataStructureDFG:
            _, post_sets = get_transitive_relations(dfg)
            efg = get_efg(post_sets)
        else:
            raise NotImplementedError

        po = generate_initial_order(alphabet, efg)
        clustered_po = cluster_order(po)

        start_activities = set(list(dfg.start_activities.keys()))
        end_activities = set(list(dfg.end_activities.keys()))
        if is_valid_order(clustered_po, efg, start_activities, end_activities):
            return clustered_po
        else:
            return None

    @classmethod
    def apply(
        cls, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[Tuple[StrictPartialOrder, List[POWL]]]:
        g = cls.holds(obj, parameters)
        if g is None:
            return g
        children = cls.project(obj, g.nodes, parameters)
        po = StrictPartialOrder(children)
        for i, j in combinations(range(len(g.nodes)), 2):
            if g.is_edge_id(i, j):
                po.order.add_edge(children[i], children[j])
            elif g.is_edge_id(j, i):
                po.order.add_edge(children[j], children[i])
        return po, po.children


def project_on_groups_with_unique_activities(
    log: Counter, groups: List[Collection[Any]]
):
    r = list()
    for g in groups:
        new_log = Counter()
        for var, freq in log.items():
            new_var = []
            for activity in var:
                if activity in g:
                    new_var.append(activity)
            new_var_tuple = tuple(new_var)
            if new_var_tuple in new_log.keys():
                new_log[new_var_tuple] = new_log[new_var_tuple] + freq
            else:
                new_log[new_var_tuple] = freq
        r.append(new_log)
    return list(map(lambda l: IMDataStructureUVCL(l), r))


class MaximalPartialOrderCutUVCL(MaximalPartialOrderCut[IMDataStructureUVCL]):
    @classmethod
    def project(
        cls,
        obj: IMDataStructureUVCL,
        groups: List[Collection[Any]],
        parameters: Optional[Dict[str, Any]] = None,
    ) -> List[IMDataStructureUVCL]:
        return project_on_groups_with_unique_activities(obj.data_structure, groups)


class MaximalPartialOrderCutDFG(MaximalPartialOrderCut[IMDataStructureDFG]):
    @classmethod
    def project(
        cls,
        obj: IMDataStructureDFG,
        groups: List[Collection[Any]],
        parameters: Optional[Dict[str, Any]] = None,
    ) -> List[IMDataStructureDFG]:
        """
        For each group g:
          - Keep only arcs (a,b) with a,b in g.
          - Start counts in the projected DFG for activity x in g:
                original_start[x] + sum_{y not in g} count(y -> x)
          - End counts in the projected DFG for activity x in g:
                original_end[x] + sum_{y not in g} count(x -> y)
          - 'skip' is set to False (all blocks are required in a partial-order composition).
        """
        base = obj.dfg

        act2grp: Dict[Any, int] = {}
        for gi, g in enumerate(groups):
            for a in g:
                act2grp[a] = gi

        dfgs: List[DFG] = [DFG() for _ in groups]
        starts_from_out: List[Counter] = [
            Counter() for _ in groups
        ]  # counts of y->x with y outside group
        ends_to_out: List[Counter] = [
            Counter() for _ in groups
        ]  # counts of x->y with y outside group

        # internal arcs and cross-boundary starts/ends
        for (a, b), count in base.graph.items():
            gi = act2grp[a]
            gj = act2grp[b]
            if gi == gj:
                dfgs[gi].graph[(a, b)] = count
            else:
                starts_from_out[gj][b] += count
                ends_to_out[gi][a] += count

        # start/end activities inside each projected DFG
        for gi, g in enumerate(groups):
            dfi = dfgs[gi]
            for a in g:
                start_count = starts_from_out[gi][a] + base.start_activities.get(a, 0)
                end_count = ends_to_out[gi][a] + base.end_activities.get(a, 0)
                if start_count > 0:
                    dfi.start_activities[a] = start_count
                if end_count > 0:
                    dfi.end_activities[a] = end_count

        return [
            IMDataStructureDFG(InductiveDFG(dfg=dfgs[i], skip=False))
            for i in range(len(dfgs))
        ]



from powl.discovery.partial_order_based.utils.simplified_objects import Graph


def combine_orders(orders):
    nodes = sorted({x for g in orders for x in g.nodes})
    idx = {x: i for i, x in enumerate(nodes)}
    n = len(nodes)

    edges_bits = [0] * n
    conflicts_bits = [0] * n
    for g in orders:
        for u in g.nodes:
            ui = idx[u]
            for v in g.nodes:
                vi = idx[v]
                if u == v:
                    conflicts_bits[ui] |= 1 << vi
                if (u, v) in g.edges:
                    edges_bits[ui] |= 1 << vi
                else:
                    conflicts_bits[ui] |= 1 << vi

    for i in range(n):
        edges_bits[i] &= ~conflicts_bits[i]

    # Transitive closure with Floyd–Warshall algorithm
    for k in range(n):
        bk = edges_bits[k]
        for i in range(n):
            if (edges_bits[i] >> k) & 1:
                # add all k→* bits into i→*
                edges_bits[i] |= bk

    # remove any direct conflict edges (they may have been added via closure)
    for i in range(n):
        edges_bits[i] &= ~conflicts_bits[i]

    # build reverse‐lookup bitsets so we can prune “middle” edges fast
    rev_bits = [0] * n
    for i in range(n):
        b = edges_bits[i]
        while b:
            lb = b & -b
            j = lb.bit_length() - 1
            rev_bits[j] |= 1 << i
            b &= ~lb

    # for each conflict (i→k), remove every j→k where i→j
    for i in range(n):
        cb = conflicts_bits[i]
        while cb:
            lb = cb & -cb
            k = lb.bit_length() - 1
            mids = edges_bits[i] & rev_bits[k]
            while mids:
                mlb = mids & -mids
                j = mlb.bit_length() - 1
                edges_bits[j] &= ~(1 << k)
                rev_bits[k] &= ~(1 << j)
                mids &= ~mlb
            cb &= ~lb

    final = {
        (nodes[i], nodes[j])
        for i in range(n)
        for j in range(n)
        if (edges_bits[i] >> j) & 1
    }

    return Graph(frozenset(nodes), frozenset(final))



from typing import Any, Dict, List, Optional, Tuple, Type

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL

from powl.discovery.total_order_based.inductive.cuts.concurrency import (
    POWLConcurrencyCutUVCL,
)
from powl.discovery.total_order_based.inductive.cuts.factory import CutFactory, S, T
from powl.discovery.total_order_based.inductive.cuts.loop import POWLLoopCutUVCL
from powl.discovery.total_order_based.inductive.cuts.sequence import (
    POWLStrictSequenceCutUVCL,
)
from powl.discovery.total_order_based.inductive.cuts.xor import (
    POWLExclusiveChoiceCutUVCL,
)
from powl.discovery.total_order_based.inductive.variants.brute_force.bf_partial_order_cut import (
    BruteForcePartialOrderCutUVCL,
)
from powl.objects.obj import POWL


class CutFactoryPOWLBruteForce(CutFactory):
    @classmethod
    def get_cuts(
        cls, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> List[Type[S]]:
        if type(obj) is IMDataStructureUVCL:
            return [
                POWLExclusiveChoiceCutUVCL,
                POWLStrictSequenceCutUVCL,
                POWLConcurrencyCutUVCL,
                POWLLoopCutUVCL,
                BruteForcePartialOrderCutUVCL,
            ]
        return list()

    @classmethod
    def find_cut(
        cls, obj: IMDataStructureUVCL, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[Tuple[POWL, List[T]]]:
        for c in CutFactoryPOWLBruteForce.get_cuts(obj):
            r = c.apply(obj, parameters)
            if r is not None:
                return r
        return None



# this file was initially copied from https://github.com/Nik314/DF2-Miner


class LeafNode:
    def __init__(self, activity, related, divergent, convergent, deficient):
        self.activity = activity
        self.related = related
        self.divergent = divergent
        self.convergent = convergent
        self.deficient = deficient

    def __str__(self, depth=0):
        indent = ""
        for i in range(0, depth):
            indent += "\t"

        result = indent + self.activity + "\n"
        result += indent + "\t Related Types: " + str(self.related) + "\n"
        result += indent + "\t Divergent Types: " + str(self.divergent) + "\n"
        result += indent + "\t Convergent Types: " + str(self.convergent) + "\n"
        result += indent + "\t Deficient Types: " + str(self.deficient) + "\n"
        return result

    def get_as_dict(self):
        return {
            "activity": self.activity,
            "related": self.related,
            "divergent": self.divergent,
            "convergent": self.convergent,
            "deficient": self.deficient,
        }

    def get_type_information(self):
        return {
            (self.activity, "rel"): self.related,
            (self.activity, "div"): self.divergent,
            (self.activity, "con"): self.convergent,
            (self.activity, "def"): self.deficient,
        }

    def get_object_types(self):
        return set(
            sum([list(value) for value in self.get_type_information().values()], [])
        )

    def get_activities(self):
        return {self.activity}

    def get_unique_relations(self):
        return set()


class OperatorNode:
    def __init__(self, operator, subtrees):
        self.operator = operator
        self.subtrees = subtrees

    def __str__(self, depth=0):
        indent = ""
        for i in range(0, depth):
            indent += "\t"
        result = indent + str(self.operator) + "\n"
        for tree in self.subtrees:
            result += tree.__str__(depth + 1)
        return result

    def get_as_dict(self):
        return {
            "operator": str(self.operator),
            "subtrees": [subtree.get_as_dict() for subtree in self.subtrees],
        }

    def get_type_information(self):
        return {
            key: value
            for subtree in self.subtrees
            for key, value in subtree.get_type_information().items()
        }

    def get_object_types(self):
        return set(
            sum([list(value) for value in self.get_type_information().values()], [])
        )

    def get_activities(self):
        return set(sum([[key[0]] for key in self.get_type_information().keys()], []))

    def get_unique_relations(self):
        if len(self.subtrees) == 1:
            return self.subtrees[0].get_unique_relations() | {self.operator}
        else:
            return set(
                sum([list(sub.get_unique_relations()) for sub in self.subtrees], [])
            )


def load_from_pt(process_tree, related, divergence, convergence, deficiency):

    if process_tree.children:
        return OperatorNode(
            process_tree.operator,
            [
                load_from_pt(sub, related, divergence, convergence, deficiency)
                for sub in process_tree.children
            ],
        )
    elif process_tree.label:
        activity = process_tree.label
        return LeafNode(
            activity,
            related[activity],
            divergence[activity],
            convergence[activity],
            deficiency[activity],
        )
    else:
        all_types = set(sum([list(v) for v in related.values()], []))
        return LeafNode("", all_types, all_types, all_types, all_types)



from typing import Any, Dict, List, Optional

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL

from powl.discovery.total_order_based.inductive.variants.decision_graph.cyclic_dg_cut import (
    CyclicDecisionGraphCutUVCL,
)


class DFGCutUVCL(CyclicDecisionGraphCutUVCL):
    @classmethod
    def holds(
        cls,
        obj: IMDataStructureUVCL,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> Optional[List[Any]]:
        alphabet = parameters["alphabet"]
        print("alphabet: ", alphabet)
        return [frozenset([a]) for a in alphabet]



from typing import Any, Dict, List, Optional, Tuple, Type

from pm4py.algo.discovery.inductive.dtypes.im_ds import (
    IMDataStructureDFG,
    IMDataStructureUVCL,
)
from pm4py.objects.dfg import util as dfu

from powl.discovery.total_order_based.inductive.cuts.concurrency import (
    POWLConcurrencyCutDFG,
    POWLConcurrencyCutUVCL,
)
from powl.discovery.total_order_based.inductive.cuts.factory import CutFactory, S, T
from powl.discovery.total_order_based.inductive.cuts.loop import (
    POWLLoopCutDFG,
    POWLLoopCutUVCL,
)
from powl.discovery.total_order_based.inductive.cuts.sequence import (
    POWLStrictSequenceCutDFG,
    POWLStrictSequenceCutUVCL,
)
from powl.discovery.total_order_based.inductive.cuts.xor import (
    POWLExclusiveChoiceCutDFG,
    POWLExclusiveChoiceCutUVCL,
)
from powl.discovery.total_order_based.inductive.variants.maximal.maximal_partial_order_cut import (
    MaximalPartialOrderCutDFG,
    MaximalPartialOrderCutUVCL,
)
from powl.objects.obj import POWL


class CutFactoryPOWLMaximal(CutFactory):
    @classmethod
    def get_cuts(
        cls, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> List[Type[S]]:
        if type(obj) is IMDataStructureUVCL:
            return [
                POWLExclusiveChoiceCutUVCL,
                POWLStrictSequenceCutUVCL,
                POWLConcurrencyCutUVCL,
                POWLLoopCutUVCL,
                MaximalPartialOrderCutUVCL,
            ]
        elif type(obj) is IMDataStructureDFG:
            return [
                POWLExclusiveChoiceCutDFG,
                POWLStrictSequenceCutDFG,
                POWLConcurrencyCutDFG,
                POWLLoopCutDFG,
                MaximalPartialOrderCutDFG,
            ]
        else:
            return []

    @classmethod
    def find_cut(
        cls, obj: IMDataStructureUVCL, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[Tuple[POWL, List[T]]]:

        alphabet = sorted(dfu.get_vertices(obj.dfg), key=lambda g: g.__str__())
        if len(alphabet) < 2:
            return None
        for c in CutFactoryPOWLMaximal.get_cuts(obj):
            r = c.apply(obj, parameters)
            if r is not None:
                return r
        return None



from abc import ABC
from typing import Any, Dict, Generic, Optional

from pm4py.algo.discovery.inductive.cuts.concurrency import (
    ConcurrencyCut,
    ConcurrencyCutDFG,
    ConcurrencyCutUVCL,
    T,
)
from pm4py.algo.discovery.inductive.dtypes.im_ds import (
    IMDataStructureDFG,
    IMDataStructureUVCL,
)

from powl.objects.obj import StrictPartialOrder


class POWLConcurrencyCut(ConcurrencyCut, ABC, Generic[T]):
    @classmethod
    def operator(
        cls, parameters: Optional[Dict[str, Any]] = None
    ) -> StrictPartialOrder:
        return StrictPartialOrder([])


class POWLConcurrencyCutUVCL(
    ConcurrencyCutUVCL, POWLConcurrencyCut[IMDataStructureUVCL]
):
    pass


class POWLConcurrencyCutDFG(ConcurrencyCutDFG, POWLConcurrencyCut[IMDataStructureDFG]):
    pass



from copy import deepcopy
from enum import Enum

import graphviz
from pm4py.objects.dfg.obj import DFG
from pm4py.util import exec_utils
from pm4py.visualization.common import gview, save as gsave

from powl.visualization.dfg.variants import base


class Variants(Enum):
    BASE = base


DEFAULT_VARIANT = Variants.BASE


def apply(dfg: DFG, variant=DEFAULT_VARIANT) -> graphviz.Digraph:
    """
    Visualize a frequency/performance directly-follows graph

    Parameters
    -----------------
    dfg
        Directly-follows graph

    Returns
    -----------------
    gviz
        Graphviz digraph
    """
    dfg_obj = deepcopy(dfg)
    return exec_utils.get_variant(variant).apply(dfg_obj)


def save(gviz, output_file_path, parameters=None):
    """
    Save the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    output_file_path
        Path where the GraphViz output should be saved
    """
    gsave.save(gviz, output_file_path, parameters=parameters)
    return ""


def view(gviz, parameters=None):
    """
    View the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    """
    return gview.view(gviz, parameters=parameters)


def matplotlib_view(gviz, parameters=None):
    """
    Views the diagram using Matplotlib

    Parameters
    ---------------
    gviz
        Graphviz
    """

    return gview.matplotlib_view(gviz, parameters=parameters)



# this file was initially copied from https://github.com/Nik314/DF2-Miner

from collections import Counter

import pm4py
from pm4py.objects.dfg.obj import DFG

from powl.objects.utils.relation import get_transitive_closure_from_counter


# def _partitions_from_edges(nodes, edges):
#     edges = [tuple(edge) for edge in edges]
#     parts = [{a} for a in nodes]
#
#     def find_group(a):
#         for g in parts:
#             if a in g:
#                 return g
#         return None
#
#     for u, v in edges:
#         gu = find_group(u)
#         gv = find_group(v)
#         assert gu is not None and gv is not None
#         if gu is gv:
#             continue
#         gu |= gv
#         parts.remove(gv)
#
#     covered = set().union(*parts) if parts else set()
#     nodes_set = set(nodes)
#     assert covered == nodes_set, "Invalid partition: missing/extra nodes!"
#     assert all(len(g) > 0 for g in parts), "Invalid partition: empty group encountered!"
#     assert sum(len(g) for g in parts) == len(covered), "Invalid partition: overlapping groups!"
#
#     return parts


def get_divergence_free_graph(relations, div, rel):

    alphabet = rel.keys()

    global_dfg = {(a, b): 0 for a in alphabet for b in alphabet}
    global_start = {a: 0 for a in alphabet}
    global_end = {a: 0 for a in alphabet}

    non_starts = []
    non_ends = []

    divergence_matrices = {}

    for ot in relations["ocel:type"].unique():

        sub_log = relations[relations["ocel:type"] == ot]
        local_dfg, local_start, local_end = pm4py.discover_dfg(
            sub_log, "ocel:activity", "ocel:timestamp", "ocel:oid"
        )

        filtered_edges = set()

        for key, value in local_dfg.items():
            if not value:
                continue
            source = key[0]
            target = key[1]
            if (ot in div[source]) and (ot in div[target]):
                if source != target:
                    filtered_edges.add(frozenset((source, target)))
            else:
                global_dfg[(source, target)] += value

        for key, value in local_start.items():
            if value:
                global_start[key] += value

        for key, value in local_end.items():
            if value:
                global_end[key] += value

        non_starts += [
            a for a in alphabet if not local_start.get(a, 0) and ot in rel[a]
        ]
        non_ends += [a for a in alphabet if not local_end.get(a, 0) and ot in rel[a]]

        divergence_matrices[ot] = filtered_edges

    global_dfg = {key: value for key, value in global_dfg.items() if value}
    global_start = {
        key: value
        for key, value in global_start.items()
        if value and key not in non_starts
    }
    global_end = {
        key: value for key, value in global_end.items() if value and key not in non_ends
    }

    closure = get_transitive_closure_from_counter(Counter(global_dfg))

    for node in alphabet:
        if node not in closure.keys():
            closure[node] = set()

    new_starts = set()
    new_ends = set()

    for node in alphabet:
        if node not in global_start.keys():
            reachable_from_start = any(
                node in closure[start] for start in global_start.keys()
            )
            if not reachable_from_start:
                new_starts.add(node)
        if node not in global_end.keys():
            reachable_from_end = any(end in closure[node] for end in global_end.keys())
            if not reachable_from_end:
                new_ends.add(node)

    for node in new_starts:
        global_start[node] = 1

    for node in new_ends:
        global_end[node] = 1

    return DFG(global_dfg, global_start, global_end), divergence_matrices



# this file was initially copied from https://github.com/Nik314/DF2-Miner

import pm4py
from pm4py.objects.process_tree.obj import Operator, ProcessTree
from pm4py.objects.process_tree.utils import generic as pt_util

from powl.discovery.object_centric.variants.oc_powl.utils.ocpn_conversion import (
    clone_workflow_net,
)
from powl.discovery.object_centric.variants.oc_tree.utils.oc_process_trees import (
    LeafNode,
    OperatorNode,
)


def project_ocpt(ocpt, object_type):

    if isinstance(ocpt, LeafNode):
        if ocpt.activity == "" or object_type not in ocpt.related:
            return ProcessTree()
        return ProcessTree(label=ocpt.activity)

    assert isinstance(ocpt, OperatorNode)
    related_activities = set(
        [
            a
            for a in ocpt.get_activities()
            if object_type in ocpt.get_type_information()[(a, "rel")] and a != ""
        ]
    )

    if not related_activities:
        return ProcessTree()

    if all(
        object_type in ocpt.get_type_information()[(a, "div")]
        for a in related_activities
    ):
        return ProcessTree(
            operator=pm4py.objects.process_tree.obj.Operator.LOOP,
            children=[
                ProcessTree(),
                ProcessTree(
                    operator=pm4py.objects.process_tree.obj.Operator.XOR,
                    children=[ProcessTree(label=a) for a in related_activities],
                ),
            ],
        )

    else:
        if ocpt.operator == Operator.PARALLEL or ocpt.operator == Operator.LOOP:
            return ProcessTree(
                operator=ocpt.operator,
                children=[project_ocpt(sub, object_type) for sub in ocpt.subtrees],
            )

        diverging = [
            i
            for i in range(len(ocpt.subtrees))
            if ocpt.subtrees[i].get_activities() & related_activities
            and all(
                object_type in ocpt.get_type_information()[(a, "div")]
                for a in ocpt.subtrees[i].get_activities() & related_activities
            )
        ]
        non_diverging = [
            i
            for i in range(len(ocpt.subtrees))
            if ocpt.subtrees[i].get_activities() & related_activities
            and i not in diverging
        ]
        skipped = [
            i
            for i in range(0, len(ocpt.subtrees))
            if i not in diverging and i not in non_diverging
        ]

        if ocpt.operator == Operator.SEQUENCE:

            children, index = [], 0

            while index < len(ocpt.subtrees):

                if index in diverging:

                    div_activities = (
                        ocpt.subtrees[index].get_activities() & related_activities
                    )
                    while index + 1 in diverging and index + 1 < len(ocpt.subtrees):
                        index += 1
                        if index not in skipped:
                            div_activities |= (
                                ocpt.subtrees[index].get_activities()
                                & related_activities
                            )

                    div_activities = {a for a in div_activities if a != ""}
                    div_subtree = ProcessTree(
                        operator=pm4py.objects.process_tree.obj.Operator.LOOP,
                        children=[
                            ProcessTree(),
                            ProcessTree(
                                operator=pm4py.objects.process_tree.obj.Operator.XOR,
                                children=[ProcessTree(label=a) for a in div_activities],
                            ),
                        ],
                    )
                    children.append(div_subtree)

                else:
                    children.append(project_ocpt(ocpt.subtrees[index], object_type))
                index += 1
            return ProcessTree(operator=Operator.SEQUENCE, children=children)

        if ocpt.operator == Operator.XOR:

            div_activities = set(
                sum(
                    [
                        list(ocpt.subtrees[i].get_activities() & related_activities)
                        for i in diverging
                    ],
                    [],
                )
            )
            div_activities = {a for a in div_activities if a != ""}
            optional = any(
                [
                    isinstance(sub, LeafNode)
                    and sub.activity == ""
                    and object_type in sub.related
                    for sub in ocpt.subtrees
                ]
            )

            if div_activities:
                div_subtree = ProcessTree(
                    operator=pm4py.objects.process_tree.obj.Operator.LOOP,
                    children=[
                        ProcessTree(),
                        ProcessTree(
                            operator=pm4py.objects.process_tree.obj.Operator.XOR,
                            children=[ProcessTree(label=a) for a in div_activities],
                        ),
                    ],
                )

                return ProcessTree(
                    operator=Operator.XOR,
                    children=[div_subtree]
                    + [
                        project_ocpt(ocpt.subtrees[i], object_type)
                        for i in non_diverging
                    ]
                    + ([ProcessTree()] if optional else []),
                )
            else:
                return ProcessTree(
                    operator=Operator.XOR,
                    children=[
                        project_ocpt(ocpt.subtrees[i], object_type)
                        for i in non_diverging
                    ]
                    + ([ProcessTree()] if optional else []),
                )


def handle_deficiency(ocpt):

    if isinstance(ocpt, OperatorNode):
        subresults = [handle_deficiency(sub) for sub in ocpt.subtrees]
        return OperatorNode(ocpt.operator, [sub[0] for sub in subresults]), sum(
            [sub[1] for sub in subresults], []
        )
    elif ocpt.activity == "":
        return ocpt, []
    else:
        from itertools import chain, combinations

        stable_types = ocpt.related - ocpt.deficient
        variable_types = ocpt.related & ocpt.deficient
        if variable_types:
            ot_sets = [
                stable_types | {c for c in comb}
                for comb in chain.from_iterable(
                    combinations(variable_types, n)
                    for n in range(len(variable_types) + 1)
                )
            ]

            children = [
                LeafNode(
                    activity=ocpt.activity + "<|>" + str(sorted(list(ots))),
                    related=ots,
                    convergent=ocpt.convergent & ots,
                    deficient=set(),
                    divergent=ocpt.divergent & ots,
                )
                for ots in ot_sets
            ]
            return OperatorNode(operator=Operator.XOR, subtrees=children), [
                ocpt.activity
            ]
        else:
            return ocpt, []


def convert_ocpt_to_ocpn(ocpt):

    assert isinstance(ocpt, OperatorNode) or isinstance(ocpt, LeafNode)
    nets = {}
    nets_duplicates = {}
    convergent_activities = {}
    activities = set()

    ocpt, special_activities = handle_deficiency(ocpt)

    for ot in ocpt.get_object_types():
        pt = project_ocpt(ocpt, ot)
        pt = pt_util.reduce_tau_leafs(pt)
        pt = pt_util.fold(pt)
        net, im, fm = pm4py.convert_to_petri_net(pt)
        nets[ot] = net, im, fm
        nets_duplicates[ot] = clone_workflow_net(net, im, fm, label_delimiter="<|>")
        activities.update(
            {
                a
                for a in ocpt.get_activities()
                if ot in ocpt.get_type_information()[(a, "rel")]
            }
        )
        convergent_activities[ot] = [
            a
            for a in ocpt.get_activities()
            if ot in ocpt.get_type_information()[(a, "con")]
        ]

    ocpn = {
        "activities": activities,
        "object_types": nets.keys(),
        "petri_nets": nets,
        "petri_nets_with_duplicates": nets_duplicates,
        "double_arcs_on_activity": convergent_activities,
        "tbr_results": {},
    }
    return ocpn



from copy import copy
from multiprocessing import Manager, Pool
from typing import Any, Dict, List, Optional, Tuple

from pm4py.algo.discovery.inductive.dtypes.im_dfg import InductiveDFG

from pm4py.algo.discovery.inductive.dtypes.im_ds import (
    IMDataStructureDFG,
    IMDataStructureUVCL,
)
from pm4py.algo.discovery.inductive.fall_through.empty_traces import (
    EmptyTracesDFG,
    EmptyTracesUVCL,
)

from powl.objects.BinaryRelation import BinaryRelation
from powl.objects.obj import DecisionGraph


class POWLEmptyTracesDecisionGraphUVCL(EmptyTracesUVCL):
    @classmethod
    def apply(
        cls,
        obj: IMDataStructureUVCL,
        pool: Pool = None,
        manager: Manager = None,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> Optional[Tuple[DecisionGraph, List[IMDataStructureUVCL]]]:
        if cls.holds(obj, parameters):
            data_structure = copy(obj.data_structure)
            del data_structure[()]
            children = [IMDataStructureUVCL(data_structure)]
            dg = DecisionGraph(
                BinaryRelation(copy(children)), children, children, empty_path=True
            )
            return dg, children
        else:
            return None


class POWLEmptyTracesDecisionGraphDFG(EmptyTracesDFG):
    @classmethod
    def apply(
        cls,
        obj: IMDataStructureDFG,
        pool=None,
        manager=None,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> Optional[Tuple[DecisionGraph, List[IMDataStructureDFG]]]:
        if cls.holds(obj, parameters):
            children = [IMDataStructureDFG(InductiveDFG(obj.data_structure.dfg))]
            dg = DecisionGraph(
                BinaryRelation(copy(children)), children, children, empty_path=True
            )
            return dg, children
        return None



from typing import Any, Dict, Optional, Type, Union

import pandas as pd
from pm4py import util
from pm4py.algo.discovery.inductive.algorithm import Parameters
from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from pm4py.objects.log.obj import EventLog
from pm4py.util import exec_utils, xes_constants as xes_util
from pm4py.util.compression import util as comut
from pm4py.util.compression.dtypes import UVCL

from powl.discovery.total_order_based.inductive.variants.im_brute_force import (
    POWLInductiveMinerBruteForce,
)
from powl.discovery.total_order_based.inductive.variants.im_decision_graph_clustering import (
    POWLInductiveMinerDecisionGraphClustering,
)

from powl.discovery.total_order_based.inductive.variants.im_decision_graph_cyclic import (
    POWLInductiveMinerDecisionGraphCyclic,
    POWLInductiveMinerDecisionGraphCyclicStrict,
)
from powl.discovery.total_order_based.inductive.variants.im_decision_graph_maximal import (
    POWLInductiveMinerDecisionGraphMaximal,
)
from powl.discovery.total_order_based.inductive.variants.im_dynamic_clustering_frequencies import (
    POWLInductiveMinerDynamicClusteringFrequency,
)
from powl.discovery.total_order_based.inductive.variants.im_maximal import (
    POWLInductiveMinerMaximalOrder,
)
from powl.discovery.total_order_based.inductive.variants.im_tree import IMBasePOWL
from powl.discovery.total_order_based.inductive.variants.powl_discovery_varaints import (
    POWLDiscoveryVariant,
)
from powl.objects.obj import POWL

DEFAULT_POWL_MINER = POWLDiscoveryVariant.DECISION_GRAPH_CYCLIC


def get_variant(variant: POWLDiscoveryVariant) -> Type[IMBasePOWL]:
    if variant == POWLDiscoveryVariant.TREE:
        return IMBasePOWL
    elif variant == POWLDiscoveryVariant.BRUTE_FORCE:
        return POWLInductiveMinerBruteForce
    elif variant == POWLDiscoveryVariant.MAXIMAL:
        return POWLInductiveMinerMaximalOrder
    elif variant == POWLDiscoveryVariant.DYNAMIC_CLUSTERING:
        return POWLInductiveMinerDynamicClusteringFrequency
    elif variant == POWLDiscoveryVariant.DECISION_GRAPH_MAX:
        return POWLInductiveMinerDecisionGraphMaximal
    elif variant == POWLDiscoveryVariant.DECISION_GRAPH_CLUSTERING:
        return POWLInductiveMinerDecisionGraphClustering
    elif variant == POWLDiscoveryVariant.DECISION_GRAPH_CYCLIC:
        return POWLInductiveMinerDecisionGraphCyclic
    elif variant == POWLDiscoveryVariant.DECISION_GRAPH_CYCLIC_STRICT:
        return POWLInductiveMinerDecisionGraphCyclicStrict
    else:
        raise Exception("Invalid Variant!")


def apply(
    obj: Union[EventLog, pd.DataFrame, UVCL],
    parameters: Optional[Dict[Any, Any]] = None,
    variant=DEFAULT_POWL_MINER,
    simplify=True,
) -> POWL:
    if parameters is None:
        parameters = {}
    ack = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_util.DEFAULT_NAME_KEY
    )
    tk = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, xes_util.DEFAULT_TIMESTAMP_KEY
    )
    cidk = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, util.constants.CASE_CONCEPT_NAME
    )
    if type(obj) in [EventLog, pd.DataFrame]:
        uvcl = comut.get_variants(
            comut.project_univariate(
                obj, key=ack, df_glue=cidk, df_sorting_criterion_key=tk
            )
        )
    else:
        uvcl = obj

    algorithm = get_variant(variant)
    im = algorithm(parameters)
    res = im.apply(IMDataStructureUVCL(uvcl), parameters)
    if simplify:
        res = res.simplify()

    return res



from multiprocessing import Manager, Pool
from typing import Any, Dict, List, Optional, Tuple

from pm4py.algo.discovery.inductive.dtypes.im_dfg import InductiveDFG

from pm4py.algo.discovery.inductive.dtypes.im_ds import (
    IMDataStructureDFG,
    IMDataStructureUVCL,
)
from pm4py.algo.discovery.inductive.fall_through.flower import (
    FlowerModelDFG,
    FlowerModelUVCL,
)
from pm4py.objects.dfg.obj import DFG
from pm4py.objects.process_tree.obj import Operator
from pm4py.util.compression import util as comut
from pm4py.util.compression.dtypes import UVCL

from powl.objects.obj import OperatorPOWL


class POWLFlowerModelUVCL(FlowerModelUVCL):
    @classmethod
    def apply(
        cls,
        obj: IMDataStructureUVCL,
        pool: Pool = None,
        manager: Manager = None,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> Optional[Tuple[OperatorPOWL, List[IMDataStructureUVCL]]]:
        log = obj.data_structure
        uvcl_redo = UVCL()
        for a in sorted(list(comut.get_alphabet(log))):
            uvcl_redo[(a,)] = 1
        uvcl_do = UVCL()
        im_uvcl_do = IMDataStructureUVCL(uvcl_do)
        im_uvcl_redo = IMDataStructureUVCL(uvcl_redo)
        return OperatorPOWL(Operator.LOOP, []), [im_uvcl_do, im_uvcl_redo]


class POWLFlowerModelDFG(FlowerModelDFG):
    @classmethod
    def apply(
        cls,
        obj: IMDataStructureDFG,
        pool: Pool = None,
        manager: Manager = None,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> Optional[Tuple[OperatorPOWL, List[IMDataStructureDFG]]]:
        activities = (
            set(obj.dfg.start_activities)
            .union(set(obj.dfg.end_activities))
            .union(set(x[0] for x in obj.dfg.graph))
            .union(set(x[1] for x in obj.dfg.graph))
        )
        dfg_redo = DFG()
        for a in activities:
            dfg_redo.start_activities[a] = 1
            dfg_redo.end_activities[a] = 1
        dfg_do = DFG()
        im_dfg_do = IMDataStructureDFG(InductiveDFG(dfg_do))
        im_dfg_redo = IMDataStructureDFG(InductiveDFG(dfg_redo))
        return OperatorPOWL(Operator.LOOP, []), [im_dfg_do, im_dfg_redo]



from collections import Counter
from typing import Any, Dict, Optional

from pm4py.algo.discovery.inductive.fall_through.tau_loop import TauLoopUVCL
from pm4py.util.compression import util as comut
from pm4py.util.compression.dtypes import UVCL

from powl.discovery.total_order_based.inductive.fall_through.strict_tau_loop import (
    POWLStrictTauLoopUVCL,
)


class POWLTauLoopUVCL(POWLStrictTauLoopUVCL, TauLoopUVCL):
    @classmethod
    def _get_projected_log(
        cls, log: UVCL, parameters: Optional[Dict[str, Any]] = None
    ) -> UVCL:
        start_activities = comut.get_start_activities(log)
        proj = Counter()
        for t in log:
            x = 0
            for i in range(1, len(t)):
                if t[i] in start_activities:
                    proj.update({t[x:i]: log[t]})
                    x = i
            proj.update({t[x : len(t)]: log[t]})
        return proj



from powl.objects.obj import POWL, SilentTransition, Transition


class ObjectCentricPOWL:
    def __init__(self) -> None:
        self.flat_model: POWL | Transition | None = None


class LeafNode(ObjectCentricPOWL):
    def __init__(
        self, transition: Transition, related, divergent, convergent, deficient
    ):
        super().__init__()
        if isinstance(transition, SilentTransition):
            self.activity = ""
        else:
            self.activity = transition.label
        self.flat_model = transition
        self.related = related
        self.divergent = divergent
        self.convergent = convergent
        self.deficient = deficient

    def get_type_information(self):
        return {
            (self.activity, "rel"): self.related,
            (self.activity, "div"): self.divergent,
            (self.activity, "con"): self.convergent,
            (self.activity, "def"): self.deficient,
        }

    def get_object_types(self):
        return set(
            sum([list(value) for value in self.get_type_information().values()], [])
        )

    def get_activities(self):
        return {self.activity}


class ComplexModel(ObjectCentricPOWL):
    def __init__(self, flat_model: POWL, mapping):
        super().__init__()
        self.flat_model = flat_model
        self.mapping = mapping
        self.oc_children = [mapping[child] for child in flat_model.children]

    def get_type_information(self):
        return {
            key: value
            for oc_child in self.oc_children
            for key, value in oc_child.get_type_information().items()
        }

    def get_object_types(self):
        return set(
            sum([list(value) for value in self.get_type_information().values()], [])
        )

    def get_activities(self):
        return set(sum([[key[0]] for key in self.get_type_information().keys()], []))


def load_oc_powl(
    flat_model: POWL, related, divergence, convergence, deficiency
) -> ObjectCentricPOWL:
    if isinstance(flat_model, Transition):
        if isinstance(flat_model, SilentTransition):
            all_types = set(sum([list(v) for v in related.values()], []))
            return LeafNode(
                SilentTransition(), all_types, all_types, all_types, all_types
            )
        else:
            activity = flat_model.label
            return LeafNode(
                flat_model,
                related[activity],
                divergence[activity],
                convergence[activity],
                deficiency[activity],
            )
    else:
        oc_children = [
            load_oc_powl(sub, related, divergence, convergence, deficiency)
            for sub in flat_model.children
        ]
        mapping = {
            flat_model.children[i]: oc_children[i]
            for i in range(len(flat_model.children))
        }
        return ComplexModel(flat_model, mapping)



from typing import Any, Dict, Optional, Type

from pm4py.algo.discovery.inductive.dtypes.im_dfg import InductiveDFG
from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureDFG
from pm4py.objects.dfg.obj import DFG

from powl.discovery.dfg_based.variants.dfg_im_decision_graph_maximal import (
    DFGPOWLInductiveMinerDecisionGraphMaximal,
)
from powl.discovery.dfg_based.variants.dfg_im_maximal import (
    DFGPOWLInductiveMinerMaximalOrder,
)
from powl.discovery.dfg_based.variants.dfg_im_tree import DFGIMBasePOWL
from powl.discovery.dfg_based.variants.im_dynamic_clustering_frequencies import (
    DFGPOWLInductiveMinerDynamicClusteringFrequency,
)
from powl.discovery.total_order_based.inductive.variants.powl_discovery_varaints import (
    POWLDiscoveryVariant,
)
from powl.objects.obj import POWL


def get_variant(variant: POWLDiscoveryVariant) -> Type[DFGIMBasePOWL]:
    if variant == POWLDiscoveryVariant.TREE:
        return DFGIMBasePOWL
    elif variant == POWLDiscoveryVariant.MAXIMAL:
        return DFGPOWLInductiveMinerMaximalOrder
    elif variant == POWLDiscoveryVariant.DYNAMIC_CLUSTERING:
        return DFGPOWLInductiveMinerDynamicClusteringFrequency
    elif variant == POWLDiscoveryVariant.DECISION_GRAPH_MAX:
        return DFGPOWLInductiveMinerDecisionGraphMaximal
    else:
        raise Exception("Invalid Variant!")


def apply(
    dfg: DFG,
    parameters: Optional[Dict[Any, Any]] = None,
    variant=POWLDiscoveryVariant.MAXIMAL,
) -> POWL:
    if parameters is None:
        parameters = {}

    im_dfg = InductiveDFG(dfg=dfg, skip=False)

    algorithm = get_variant(variant)
    im = algorithm(parameters)
    res = im.apply(IMDataStructureDFG(im_dfg), parameters)
    res = res.simplify()

    return res



import base64
import copy
import os
import re
import shutil
import tempfile
import webbrowser
from enum import Enum

from pm4py.util import exec_utils

from powl.objects.obj import POWL

from powl.visualization.powl.variants import basic, net


class POWLVisualizationVariants(Enum):
    BASIC = basic
    NET = net


DEFAULT_VARIANT = POWLVisualizationVariants.BASIC


def inline_images_and_svgs(svg_content):
    img_pattern = re.compile(r'<image[^>]+xlink:href=["\'](.*?)["\'][^>]*>')

    def encode_file_to_base64(file_path):
        with open(file_path, "rb") as file:
            return base64.b64encode(file.read()).decode("utf-8")

    def read_file_content_and_viewbox(file_path):
        with open(file_path, "r") as file:
            content = file.read()
            content = re.sub(r"<\?xml.*?\?>", "", content, flags=re.DOTALL)
            content = re.sub(r"<!DOCTYPE.*?>", "", content, flags=re.DOTALL)
            viewBox_match = re.search(r'viewBox="([^"]*)"', content)
            viewBox = viewBox_match.group(1) if viewBox_match else "0 0 1 1"
            svg_content_match = re.search(r"<svg[^>]*>(.*?)</svg>", content, re.DOTALL)
            svg_content = svg_content_match.group(1) if svg_content_match else content
            return svg_content, viewBox

    def replace_with_inline_content(match):
        file_path = match.group(1)
        if file_path.lower().endswith(".svg"):
            svg_data, viewBox = read_file_content_and_viewbox(file_path)
            viewBox_values = [float(v) for v in viewBox.split()]
            actual_width, actual_height = viewBox_values[2], viewBox_values[3]

            intended_width = float(
                match.group(0).split('width="')[1].split('"')[0].replace("px", "")
            )
            intended_height = float(
                match.group(0).split('height="')[1].split('"')[0].replace("px", "")
            )
            x = float(match.group(0).split('x="')[1].split('"')[0])
            y = float(match.group(0).split('y="')[1].split('"')[0])

            scale_x = intended_width / actual_width
            scale_y = intended_height / actual_height

            return f'<g transform="translate({x},{y}) scale({scale_x},{scale_y})">{svg_data}</g>'
        else:
            base64_data = encode_file_to_base64(file_path)
            return match.group(0).replace(
                file_path, f"data:image/png;base64,{base64_data}"
            )

    return img_pattern.sub(replace_with_inline_content, svg_content)


def apply(powl: POWL, variant=DEFAULT_VARIANT, frequency_tags=True) -> str:
    """
    Method for POWL model representation

    Parameters
    -----------
    powl
        POWL model
    parameters
        Possible parameters of the algorithm:
            Parameters.FORMAT -> Format of the image (PDF, PNG, SVG; default PNG)
    variant
        Variant of the algorithm to use:
            - POWLVisualizationVariants.BASIC (default)
            - POWLVisualizationVariants.NET: BPMN-like visualization with decision gates
    frequency_tags
        Simplify the visualization using frequency tags

    Returns
    -----------
    str
        SVG Content
    """
    powl = copy.deepcopy(powl)
    if frequency_tags:
        powl = powl.simplify_using_frequent_transitions()


    viz = exec_utils.get_variant(variant).apply(powl)
    svg_content = viz.pipe().decode("utf-8")

    svg_content_with_inline_images = inline_images_and_svgs(svg_content)

    return svg_content_with_inline_images


def save(svg_content: str, output_file_path: str):
    """
    Save the diagram in a specified format.

    Parameters
    -----------
    svg_content : str
        SVG content
    output_file_path : str
        Path where the output file should be saved
    """

    with tempfile.NamedTemporaryFile(delete=False, mode="w+", suffix=".svg") as tmpfile:
        tmpfile.write(svg_content)
        tmpfile_path = tmpfile.name

    if output_file_path.endswith("svg"):
        shutil.move(tmpfile_path, output_file_path)
    else:
        raise Exception(f"Unsupported format! Please use 'svg'.")

    if os.path.exists(tmpfile_path):
        os.remove(tmpfile_path)


def view(svg_content: str):
    """
    View the diagram

    Parameters
    -----------
    svg_content
        SVG content
    """

    with tempfile.NamedTemporaryFile(delete=False, mode="w+", suffix=".svg") as tmpfile:
        tmpfile.write(svg_content)
        tmpfile_path = tmpfile.name

        absolute_path = os.path.abspath(tmpfile_path)
        return webbrowser.open("file://" + absolute_path)
